import logging
import json
import time
from typing import Optional, Literal
import queue

import numpy as np
import yarp

from wrapyfi.connect.clients import Client, Clients
from wrapyfi.middlewares.yarp import YarpMiddleware
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook


class YarpClient(Client):

    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: Literal["tcp", "udp", "mcast"] = "tcp",
        persistent: bool = True,
        yarp_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the client.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param yarp_kwargs: dict: Additional kwargs for  the Yarp middleware
        :param kwargs: dict: Additional kwargs for the client
        """
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        YarpMiddleware.activate(**yarp_kwargs or {})

        self.persistent = persistent

    def close(self):
        """
        Close the client.
        """
        if hasattr(self, "_port") and self._port:
            if self._port is not None:
                self._port.close()

    def __del__(self):
        self.close()


@Clients.register("NativeObject", "yarp")
class YarpNativeObjectClient(YarpClient):
    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: Literal["tcp", "udp", "mcast"] = "tcp",
        persistent: bool = True,
        serializer_kwargs: Optional[dict] = None,
        deserializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The NativeObject listener using the YARP Bottle construct assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a Python object.

        :param name: str: Name of the client
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        :param kwargs: dict: Additional kwargs for the client
        """
        super().__init__(
            name, in_topic, carrier=carrier, persistent=persistent, **kwargs
        )
        self._port = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

    def establish(self):
        """
        Establish the client's connection to the YARP service.
        """
        while not yarp.Network.exists(self.in_topic):
            logging.info(f"[YARP] Waiting for input port: {self.in_topic}")
            time.sleep(0.2)
        self._port = yarp.RpcClient()
        rnd_id = str(np.random.randint(100000, size=1)[0])
        self._port.open(self.in_topic + ":in" + rnd_id)
        self._port.addOutput(self.in_topic, self.carrier)
        if self.persistent:
            self.established = True

    def request(self, *args, **kwargs):
        """
        Send a request to the YARP service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: Any: The response from the YARP service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except Exception as e:
            logging.error("[YARP] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the YARP service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        args_str = json.dumps(
            [args, kwargs],
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        )
        args_msg = yarp.Bottle()
        args_msg.clear()
        args_msg.addString(args_str)

        msg = yarp.Bottle()
        msg.clear()

        self._port.write(args_msg, msg)
        obj = json.loads(
            msg.get(0).asString(),
            object_hook=self._plugin_decoder_hook,
            **self._deserializer_kwargs,
        )
        self._queue.put(obj, block=False)

    def _await_reply(self):
        """
        Wait for and return the reply from the YARP service.

        :return: Any: The response from the YARP service
        """
        try:
            reply = self._queue.get(block=True)
            return reply
        except queue.Full:
            logging.warning(
                f"[YARP] Discarding data because queue is full. "
                f"This happened due to bad synchronization in {self.__name__}"
            )
            return None


@Clients.register("Image", "yarp")
class YarpImageClient(YarpNativeObjectClient):
    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: Literal["tcp", "udp", "mcast"] = "tcp",
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        persistent: bool = True,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The Image client using the YARP Bottle construct parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param width: int: The width of the image. Default is -1
        :param height: int: The height of the image. Default is -1
        :param rgb: bool: Whether the image is RGB. Default is True
        :param fp: bool: Whether to utilize floating-point precision. Default is False
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        if "jpg" in kwargs:
            logging.warning(
                "[YARP] YARP currently does not support JPG encoding in REQ/REP. Using raw image."
            )
            kwargs.pop("jpg")
        super().__init__(
            name,
            in_topic,
            carrier=carrier,
            persistent=persistent,
            serializer_kwargs=serializer_kwargs,
            **kwargs,
        )
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the YARP service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        args_str = json.dumps(
            [args, kwargs],
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        )
        args_msg = yarp.Bottle()
        args_msg.clear()
        args_msg.addString(args_str)
        msg = yarp.Bottle()
        msg.clear()
        self._port.write(args_msg, msg)
        img = json.loads(
            msg.get(0).asString(),
            object_hook=self._plugin_decoder_hook,
            **self._deserializer_kwargs,
        )
        height, width, channels = img.shape
        if 0 < self.width != width or 0 < self.height != height:
            raise ValueError("Incorrect image shape for client")
        else:
            self._queue.put(img, block=False)


@Clients.register("AudioChunk", "yarp")
class YarpAudioChunkClient(YarpNativeObjectClient):
    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: Literal["tcp", "udp", "mcast"] = "tcp",
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        persistent: bool = True,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The AudioChunk client using the YARP Bottle construct parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param channels: int: Number of audio channels. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: The size of audio chunks. Default is -1
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(
            name,
            in_topic,
            carrier=carrier,
            persistent=persistent,
            serializer_kwargs=serializer_kwargs,
            **kwargs,
        )
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the YARP service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        args_str = json.dumps(
            [args, kwargs],
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        )
        args_msg = yarp.Bottle()
        args_msg.clear()
        args_msg.addString(args_str)
        msg = yarp.Bottle()
        msg.clear()
        self._port.write(args_msg, msg)
        chunk, channels, rate, aud = json.loads(
            msg.get(0).asString(),
            object_hook=self._plugin_decoder_hook,
            **self._deserializer_kwargs,
        )
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for client")
        if (
            0 < self.chunk != chunk
            or self.channels != channels
            or aud.size != chunk * channels
        ):
            raise ValueError("Incorrect audio shape for client")
        else:
            self._queue.put((aud, rate), block=False)



"""
Encoder and Decoder for PyArrow StructArray Data via Wrapyfi.

This script provides mechanisms to encode and decode PyArrow StructArray Data using Wrapyfi.
It utilizes base64 encoding and pickle serialization.

The script contains a class, `PyArrowArray`, registered as a plugin to manage the
conversion of PyArrow StructArray data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - PyArrow: A cross-language development platform for in-memory data (refer to https://arrow.apache.org/install/ for installation instructions)
        Note: If PyArrow is not available, HAVE_PYARROW will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install pyarrow``  # Basic installation of PyArrow
"""

import pickle
import base64

from wrapyfi.utils import *

try:
    import pyarrow as pa

    HAVE_PYARROW = True
except ImportError:
    HAVE_PYARROW = False


@PluginRegistrar.register(
    types=None if not HAVE_PYARROW else pa.StructArray.__mro__[:-1]
)
class PyArrowArray(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the PyArrowArray plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode PyArrow StructArray data using pickle and base64.

        :param obj: pa.StructArray: The PyArrow StructArray data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, pickled data string, and any buffer data
        """
        buffers = []
        obj_data = pickle.dumps(obj, protocol=5, buffer_callback=buffers.append)
        obj_buffers = list(
            map(lambda x: base64.b64encode(memoryview(x)).decode("ascii"), buffers)
        )
        return True, dict(
            __wrapyfi__=(
                str(self.__class__.__name__),
                obj_data.decode("latin1"),
                *obj_buffers,
            )
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a pickled and base64 encoded string back into PyArrow StructArray data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the pickled data string and any buffer data
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, pa.StructArray]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - pa.StructArray: The decoded PyArrow StructArray data
        """
        obj_data = obj_full[1].encode("latin1")
        obj_buffers = list(
            map(lambda x: base64.b64decode(x.encode("ascii")), obj_full[2:])
        )
        obj_data = bytearray(obj_data)
        for buf in obj_buffers:
            obj_data += buf
        return True, pickle.loads(obj_data, buffers=obj_buffers)



"""
A message publisher and listener for ROS messages using Wrapyfi.

This script demonstrates the capability to transmit ROS messages, specifically
geometry_msgs/Pose and std_msgs/String, using the MiddlewareCommunicator within
the Wrapyfi library. The communication follows the PUB/SUB pattern allowing
message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the ROS message
    - Transmitting std_msgs/String and geometry_msgs/Pose ROS messages
    - Applying the PUB/SUB pattern with channeling and mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - ROS, ROSPy: Used for handling and creating ROS messages (refer to the Wrapyfi documentation for installation instructions)

    Ensure ROS is installed and the required message types are available.

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 ros_message_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the received ROS messages
        ``python3 ros_message_example.py --mode listen``
"""

import argparse

from geometry_msgs.msg import Pose, Quaternion
from std_msgs.msg import String

from wrapyfi.connect.wrapper import MiddlewareCommunicator


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "ROSMessage",
        "ros",
        "Notifier",
        "/notify/test_ros_msg_str_exchange",
        should_wait=True,
    )
    @MiddlewareCommunicator.register(
        "ROSMessage",
        "ros",
        "Notifier",
        "/notify/test_ros_msg_pose_exchange",
        should_wait=True,
    )
    def send_message(self):
        """
        Exchange ROS messages over ROS.
        """
        msg = input("Type your message: ")
        quat = Quaternion()
        quat.x = 0.1
        quat.y = -0.1
        quat.z = 0.8
        quat.z = -0.8
        ret_str = String(msg)
        ret_pose = Pose()
        ret_pose.orientation = quat
        return ret_str, ret_pose


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="A message publisher and listener for ROS messages using Wrapyfi."
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="publish",
        choices={"publish", "listen"},
        help="The transmission mode",
    )
    return parser.parse_args()


def main(args):
    """
    Main function to initiate Notify class and communication.
    """
    ros_message = Notifier()
    ros_message.activate_communication(Notifier.send_message, mode=args.mode)

    while True:
        my_string_message, my_pose_message = ros_message.send_message()
        if my_string_message is not None or my_pose_message is not None:
            print("Method result:", my_string_message, my_pose_message)


if __name__ == "__main__":
    args = parse_args()
    main(args)



"""
This example shows how to use the MiddlewareCommunicator to send and receive messages over Websockets. It can be used to test the
functionality of the Websockets using the PUB/SUB pattern. The example can be run on a single
machine or on multiple machines. In this example (as with all other examples), the publisher transmits a message over the topic '/hello/my_message'.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - Websockets (refer to the Wrapyfi documentation for installation instructions)

Run:
    # PUB/SUB mode - Publisher transmits message and prints the received object (assuming the websocket server is running). Only one instance of the websocket_server.py should be running

        ``python3 publisher_client.py``

"""

import argparse

from wrapyfi.connect.wrapper import MiddlewareCommunicator


class HelloWorld(MiddlewareCommunicator):

    @MiddlewareCommunicator.register(
        "NativeObject",
        "websocket",
        "HelloWorld",
        "/hello/my_message",
        carrier="tcp",
        should_wait=True,
    )
    def send_message(self, arg_from_requester=""):
        """
        Exchange messages and mirror user input.
        """
        msg = input("Type your message: ")
        obj = {"message": msg, "message_from_requester": arg_from_requester}
        return (obj,)


if __name__ == "__main__":
    hello_world = HelloWorld()
    hello_world.activate_communication(HelloWorld.send_message, mode="publish")

    while True:
        (my_message,) = hello_world.send_message(
            arg_from_requester=f"I got this message from the script running in publisher mode",
        )
        if my_message is not None:
            print("Method result:", my_message)



import logging
import atexit

import rospy
import std_msgs.msg
import sensor_msgs.msg


from wrapyfi.utils import SingletonOptimized
from wrapyfi.connect.wrapper import MiddlewareCommunicator


class ROSMiddleware(metaclass=SingletonOptimized):
    """
    ROS middleware wrapper. This class is a singleton, so it can be instantiated only once. The ``activate`` method
    should be called to initialize the middleware. The ``deinit`` method should be called to deinitialize the middleware
    and destroy all connections. The ``activate`` and ``deinit`` methods are automatically called when the class is
    instantiated and when the program exits, respectively.
    """

    @staticmethod
    def activate(**kwargs):
        """
        Activate the ROS middleware. This method should be called to initialize the middleware.

        :param kwargs: dict: Keyword arguments to be passed to the ROS initialization function
        """
        ROSMiddleware(**kwargs)

    def __init__(
        self,
        node_name: str = "wrapyfi",
        anonymous: bool = True,
        disable_signals: bool = True,
        *args,
        **kwargs,
    ):
        """
        Initialize the ROS middleware. This method is automatically called when the class is instantiated.

        :param node_name: str: The name of the ROS node
        :param anonymous: bool: Whether the ROS node should be anonymous
        :param disable_signals: bool: Whether the ROS node should disable signals
        :param args: list: Positional arguments to be passed to the ROS initialization function
        :param kwargs: dict: Keyword arguments to be passed to the ROS initialization function
        """
        logging.info("Initialising ROS middleware")
        rospy.init_node(node_name, anonymous=anonymous, disable_signals=disable_signals)
        atexit.register(MiddlewareCommunicator.close_all_instances)
        atexit.register(self.deinit)

    @staticmethod
    def deinit():
        """
        Deinitialize the ROS middleware. This method is automatically called when the program exits.
        """
        logging.info("Deinitializing ROS middleware")
        rospy.signal_shutdown("Deinit")


class ROSNativeObjectService(object):
    _type = "wrapyfi_services/ROSNativeObject"
    _md5sum = "46a550fd1ca640b396e26ebf988aed7b"  # AddTwoInts '6a2e34150c00229791cc89ff309fff21'
    _request_class = std_msgs.msg.String
    _response_class = std_msgs.msg.String


class ROSImageService(object):
    _type = "wrapyfi_services/ROSImage"
    _md5sum = "f720f2021b4bbbe86b0f93b08906381c"  # AddTwoInts '6a2e34150c00229791cc89ff309fff21'
    _request_class = std_msgs.msg.String
    _response_class = sensor_msgs.msg.Image



"""
A message publisher and listener for native Python objects and xarray DataArrays.

This script demonstrates the capability to transmit native Python objects and xarray DataArrays using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and xarray DataArrays
    - Applying the PUB/SUB pattern with mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - NumPy: Used for creating arrays (installed with Wrapyfi)
    - xarray, pandas: Used for creating and handling DataArrays

    Install using pip:
        ``pip install pandas xarray``

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 xarray_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 xarray_example.py --mode listen``
"""

import argparse

try:
    import numpy as np
    import pandas as pd
    import xarray as xr
except ImportError:
    print("Install xarray, pandas, and NumPy before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "Notifier",
        "/notify/test_xarray_exchange",
        carrier="tcp",
        should_wait=True,
    )
    def exchange_object(self, mware=None):
        """
        Exchange messages with xarray DataArrays and other native Python objects.
        """
        msg = input("Type your message: ")

        # Creating an example xarray DataArray
        data = np.random.rand(4, 3)
        locs = ["IA", "IL", "IN"]
        times = pd.date_range("2000-01-01", periods=4)
        da = xr.DataArray(
            data, coords=[times, locs], dims=["time", "space"], name="example"
        )

        ret = {
            "message": msg,
            "xarray_dataarray": da,
            "additional_info": {
                "set": {"a", 1, None},
                "list": [[[3, [4], 5.677890, 1.2]]],
            },
        }
        return (ret,)


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="A message publisher and listener for native Python objects and xarray DataArrays."
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="publish",
        choices={"publish", "listen"},
        help="The transmission mode",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission",
    )
    return parser.parse_args()


def main(args):
    """
    Main function to initiate Notifier class and communication.
    """
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        (msg_object,) = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)



"""
Encoder and Decoder for Dask Arrays/Dataframes Data via Wrapyfi.

This script provides mechanisms to encode and decode Dask data using Wrapyfi.
It utilizes base64 encoding to convert binary data into ASCII strings.

The script contains a class, `DaskData`, registered as a plugin to manage the
conversion of Dask data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - Dask: A flexible library for parallel computing in Python (refer to https://docs.dask.org/en/latest/install.html for installation instructions)
    - pandas: A data structures library for data analysis, time series, and statistics (refer to https://pandas.pydata.org/pandas-docs/stable/getting_started/install.html for installation instructions)
        Note: If Dask or pandas is not available, HAVE_DASK will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install pandas dask[complete]``  # Basic installation of Dask and pandas
"""

import io
import base64

import numpy as np

from wrapyfi.utils import *

try:
    import dask.dataframe as dd
    import dask.array as da
    import pandas as pd

    HAVE_DASK = True
except ImportError:
    HAVE_DASK = False


@PluginRegistrar.register(
    types=(
        None
        if not HAVE_DASK
        else dd.DataFrame.__mro__[:-1] + dd.Series.__mro__[:-1] + da.Array.__mro__[:-1]
    )
)
class DaskData(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the DaskData plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode Dask data into a base64 ASCII string.

        :param obj: Union[dd.DataFrame, dd.Series, da.Array]: The Dask data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, and object type
        """
        obj_type = None
        obj_partitions = 1
        with io.BytesIO() as memfile:
            if isinstance(obj, dd.DataFrame):
                pandas_df = obj.compute().reset_index()
                memfile.write(pandas_df.to_json(orient="records").encode("ascii"))
                obj_partitions = obj.npartitions
                obj_type = "DataFrame"
            elif isinstance(obj, dd.Series):
                pandas_ds = obj.compute().reset_index()
                memfile.write(pandas_ds.to_json(orient="records").encode("ascii"))
                obj_partitions = obj.npartitions
                obj_type = "Series"
            elif isinstance(obj, da.Array):
                np.save(memfile, obj.compute(), allow_pickle=True)
                obj_type = "Array"
            memfile.seek(0)
            obj_data = base64.b64encode(memfile.read()).decode("ascii")
        return True, dict(
            __wrapyfi__=(
                str(self.__class__.__name__),
                obj_data,
                obj_type,
                obj_partitions,
            )
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into Dask data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string and object type
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, Union[dd.DataFrame, dd.Series, da.Array]]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - Union[dd.DataFrame, da.Array]: The decoded Dask data
        """
        obj_data = base64.b64decode(obj_full[1].encode("ascii"))
        obj_type = obj_full[2]
        obj_partitions = obj_full[3]
        with io.BytesIO(obj_data) as memfile:
            if obj_type == "DataFrame":
                pandas_df = pd.read_json(
                    memfile.read().decode("ascii"), orient="records"
                )
                pandas_df.set_index(
                    "index", inplace=True
                )  # Set the index back after reading from JSON
                return True, dd.from_pandas(pandas_df, npartitions=obj_partitions)
            elif obj_type == "Series":
                pandas_ds = pd.read_json(
                    memfile.read().decode("ascii"), orient="records"
                )
                pandas_ds.set_index(
                    "index", inplace=True
                )  # Set the index back after reading from JSON
                return True, dd.from_pandas(pandas_ds, npartitions=obj_partitions)
            elif obj_type == "Array":
                np_array = np.load(memfile, allow_pickle=True)
                return True, da.from_array(np_array, chunks=np_array.shape)



import logging
import json
import time
import os
import threading
import base64
import io
from typing import Optional, Tuple

import numpy as np
import cv2
import zmq

from wrapyfi.connect.publishers import Publisher, Publishers, PublisherWatchDog
from wrapyfi.middlewares.zeromq import ZeroMQMiddlewarePubSub
from wrapyfi.encoders import JsonEncoder


SOCKET_IP = os.environ.get("WRAPYFI_ZEROMQ_SOCKET_IP", "127.0.0.1")
SOCKET_PUB_PORT = int(os.environ.get("WRAPYFI_ZEROMQ_SOCKET_PUB_PORT", 5555))
SOCKET_SUB_PORT = int(os.environ.get("WRAPYFI_ZEROMQ_SOCKET_SUB_PORT", 5556))
PARAM_PUB_PORT = int(os.environ.get("WRAPYFI_ZEROMQ_PARAM_PUB_PORT", 5655))
PARAM_SUB_PORT = int(os.environ.get("WRAPYFI_ZEROMQ_PARAM_SUB_PORT", 5656))
PARAM_REQREP_PORT = int(os.environ.get("WRAPYFI_ZEROMQ_PARAM_REQREP_PORT", 5659))
PARAM_POLL_INTERVAL = int(os.environ.get("WRAPYFI_ZEROMQ_PARAM_POLL_INTERVAL", 1))
START_PROXY_BROKER = (
    os.environ.get("WRAPYFI_ZEROMQ_START_PROXY_BROKER", True) != "False"
)
PROXY_BROKER_SPAWN = os.environ.get("WRAPYFI_ZEROMQ_PROXY_BROKER_SPAWN", "process")
ZEROMQ_PUBSUB_MONITOR_TOPIC = os.environ.get(
    "WRAPYFI_ZEROMQ_PUBSUB_MONITOR_TOPIC", "ZEROMQ/CONNECTIONS"
)
ZEROMQ_PUBSUB_MONITOR_LISTENER_SPAWN = os.environ.get(
    "WRAPYFI_ZEROMQ_PUBSUB_MONITOR_LISTENER_SPAWN", "process"
)
WATCHDOG_POLL_REPEAT = None


if os.name == "nt" and PROXY_BROKER_SPAWN == "process" and START_PROXY_BROKER:
    PROXY_BROKER_SPAWN = "thread"
    logging.warning(
        "[ZeroMQ] Wrapyfi does not support multiprocessing on Windows. Please set "
        "the environment variable WRAPYFI_ZEROMQ_PROXY_BROKER_SPAWN='thread'. "
        "Switching automatically to 'thread' mode. "
    )

if os.name == "nt" and ZEROMQ_PUBSUB_MONITOR_LISTENER_SPAWN == "process":
    ZEROMQ_PUBSUB_MONITOR_LISTENER_SPAWN = "thread"
    logging.warning(
        "[ZeroMQ] Wrapyfi does not support multiprocessing on Windows. Please set "
        "the environment variable WRAPYFI_ZEROMQ_PUBSUB_MONITOR_LISTENER_SPAWN='thread'. "
        "Switching automatically to 'thread' mode. "
    )


class ZeroMQPublisher(Publisher):
    def __init__(
        self,
        name: str,
        out_topic: str,
        carrier: str = "tcp",
        should_wait: bool = True,
        socket_ip: str = SOCKET_IP,
        socket_pub_port: int = SOCKET_PUB_PORT,
        socket_sub_port: int = SOCKET_SUB_PORT,
        start_proxy_broker: bool = START_PROXY_BROKER,
        proxy_broker_spawn: str = PROXY_BROKER_SPAWN,
        pubsub_monitor_topic: str = ZEROMQ_PUBSUB_MONITOR_TOPIC,
        pubsub_monitor_listener_spawn: Optional[
            str
        ] = ZEROMQ_PUBSUB_MONITOR_LISTENER_SPAWN,
        zeromq_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the publisher and start the proxy broker if necessary.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param socket_ip: str: IP address of the socket. Default is '127.0.0.1'
        :param socket_pub_port: int: Port of the socket for publishing. Default is 5555
        :param socket_sub_port: int: Port of the socket for subscribing. Default is 5556
        :param start_proxy_broker: bool: Whether to start a proxy broker. Default is True
        :param proxy_broker_spawn: str: Whether to spawn the proxy broker as a process or thread. Default is 'process'
        :param pubsub_monitor_topic: str: Topic to monitor the connections. Default is 'ZEROMQ/CONNECTIONS'
        :param pubsub_monitor_listener_spawn: str: Whether to spawn the PUB/SUB monitor listener as a process or thread. Default is 'process'
        :param zeromq_kwargs: dict: Additional kwargs for the ZeroMQ Pub/Sub middleware
        :param kwargs: Additional kwargs for the publisher
        """
        if carrier or carrier != "tcp":
            logging.warning(
                "[ZeroMQ] ZeroMQ does not support other carriers than TCP for PUB/SUB pattern. Using TCP."
            )
            carrier = "tcp"
        super().__init__(
            name, out_topic, carrier=carrier, should_wait=should_wait, **kwargs
        )

        self.socket_pub_address = f"{carrier}://{socket_ip}:{socket_pub_port}"
        self.socket_sub_address = f"{carrier}://{socket_ip}:{socket_sub_port}"

        ZeroMQMiddlewarePubSub.activate(
            socket_pub_address=self.socket_pub_address,
            socket_sub_address=self.socket_sub_address,
            start_proxy_broker=start_proxy_broker,
            proxy_broker_spawn=proxy_broker_spawn,
            pubsub_monitor_topic=pubsub_monitor_topic,
            pubsub_monitor_listener_spawn=pubsub_monitor_listener_spawn,
            **zeromq_kwargs or {},
        )

        ZeroMQMiddlewarePubSub().shared_monitor_data.add_topic(self.out_topic)

    def await_connection(
        self, out_topic: Optional[str] = None, repeats: Optional[int] = None
    ):
        """
        Wait for the connection to be established.

        :param out_topic: str: Name of the output topic
        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        connected = False
        if out_topic is None:
            out_topic = self.out_topic
        logging.info(f"[ZeroMQ] Waiting for output connection: {out_topic}")
        if repeats is None:
            if self.should_wait:
                repeats = -1
            else:
                return True
        while repeats > 0 or repeats <= -1:
            repeats -= 1
            connected = ZeroMQMiddlewarePubSub().shared_monitor_data.is_connected(
                out_topic
            )
            if connected:
                break
            time.sleep(0.02)
        logging.info(f"[ZeroMQ] Output connection established: {out_topic}")
        return connected

    def close(self):
        """
        Close the publisher.
        """
        ZeroMQMiddlewarePubSub().shared_monitor_data.remove_topic(self.out_topic)
        time.sleep(0.2)

        if hasattr(self, "_socket") and self._socket:
            if self._socket is not None:
                self._socket.close()

    def __del__(self):
        self.close()


@Publishers.register("NativeObject", "zeromq")
class ZeroMQNativeObjectPublisher(ZeroMQPublisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        carrier: str = "tcp",
        should_wait: bool = True,
        multi_threaded: bool = False,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The NativeObjectPublisher using the ZeroMQ message construct assuming a combination of python native objects
        and numpy arrays as input. Serializes the data (including plugins) using the encoder and sends it as a string.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate socket for each thread. Default is False
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param kwargs: dict: Additional kwargs for the publisher
        """
        super().__init__(
            name, out_topic, carrier=carrier, should_wait=should_wait, **kwargs
        )
        self._socket = self._netconnect = None
        if multi_threaded:
            self._thread_local_storage = threading.local()

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        self._socket = zmq.Context.instance().socket(zmq.PUB)
        for socket_property in ZeroMQMiddlewarePubSub().zeromq_kwargs.items():
            if isinstance(socket_property[1], str):
                self._socket.setsockopt_string(
                    getattr(zmq, socket_property[0]), socket_property[1]
                )
            else:
                self._socket.setsockopt(
                    getattr(zmq, socket_property[0]), socket_property[1]
                )
        self._socket.connect(self.socket_sub_address)
        self._topic = self.out_topic.encode()
        established = self.await_connection(repeats=repeats)
        return self.check_establishment(established)

    def _get_socket(self):
        """
        Retrieve or create a ZeroMQ socket specific to the current thread.
        """
        if not hasattr(self._thread_local_storage, "socket"):
            # Initialize a new socket for the thread
            self._thread_local_storage.socket = zmq.Context.instance().socket(zmq.PUB)
            for socket_property in ZeroMQMiddlewarePubSub().zeromq_kwargs.items():
                if isinstance(socket_property[1], str):
                    self._thread_local_storage.socket.setsockopt_string(
                        getattr(zmq, socket_property[0]), socket_property[1]
                    )
                else:
                    self._thread_local_storage.socket.setsockopt(
                        getattr(zmq, socket_property[0]), socket_property[1]
                    )
            self._thread_local_storage.socket.connect(self.socket_sub_address)
        return self._thread_local_storage.socket

    def publish(self, obj):
        """
        Publish the object to the middleware.

        :param obj: object: Object to publish
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)
        obj_str = json.dumps(
            obj,
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        )
        if hasattr(self, "_thread_local_storage"):
            socket = self._get_socket()
        else:
            socket = self._socket
        socket.send_multipart([self._topic, obj_str.encode()])


@Publishers.register("Image", "zeromq")
class ZeroMQImagePublisher(ZeroMQNativeObjectPublisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        carrier: str = "tcp",
        should_wait: bool = True,
        multi_threaded: bool = False,
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        jpg: bool = False,
        **kwargs,
    ):
        """
        The ImagePublisher using the ZeroMQ message construct assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate socket for each thread. Default is False
        :param width: int: Width of the image. Default is -1 meaning that the width is not fixed
        :param height: int: Height of the image. Default is -1 meaning that the height is not fixed
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be compressed as JPG. Default is False
        """
        super().__init__(
            name, out_topic, carrier=carrier, should_wait=should_wait, **kwargs
        )
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._type = np.float32 if self.fp else np.uint8

    def publish(self, img: np.ndarray):
        """
        Publish the image to the middleware.

        :param img: np.ndarray: Image to publish formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if img is None:
            return

        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)
        if (
            0 < self.width != img.shape[1]
            or 0 < self.height != img.shape[0]
            or not (
                (img.ndim == 2 and not self.rgb)
                or (img.ndim == 3 and self.rgb and img.shape[2] == 3)
            )
        ):
            raise ValueError("Incorrect image shape for publisher")
        if not img.flags["C_CONTIGUOUS"]:
            img = np.ascontiguousarray(img)

        if self.jpg:
            img_str = np.array(cv2.imencode(".jpg", img)[1]).tostring()
        else:
            img_str = json.dumps(
                img,
                cls=self._plugin_encoder,
                **self._plugin_kwargs,
                serializer_kwrags=self._serializer_kwargs,
            ).encode()
        img_header = "{timestamp:" + str(time.time()) + "}"
        if hasattr(self, "_thread_local_storage"):
            socket = self._get_socket()
        else:
            socket = self._socket
        socket.send_multipart([self._topic, img_header.encode(), img_str])


@Publishers.register("AudioChunk", "zeromq")
class ZeroMQAudioChunkPublisher(ZeroMQNativeObjectPublisher):
    def __init__(
        self,
        name: str,
        out_topic: str,
        carrier: str = "tcp",
        should_wait: bool = True,
        multi_threaded: bool = False,
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        **kwargs,
    ):
        """
        The AudioChunkPublisher using the ZeroMQ message construct assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate socket for each thread. Default is False
        :param channels: int: Number of channels. Default is 1
        :param rate: int: Sampling rate. Default is 44100
        :param chunk: int: Chunk size. Default is -1 meaning that the chunk size is not fixed
        """
        super().__init__(
            name,
            out_topic,
            carrier=carrier,
            should_wait=should_wait,
            multi_threaded=multi_threaded,
            width=chunk,
            height=channels,
            rgb=False,
            fp=True,
            jpg=False,
            **kwargs,
        )
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    def publish(self, aud: Tuple[np.ndarray, int]):
        """
        Publish the audio chunk to the middleware.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        aud, rate = aud
        if aud is None:
            return
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for publisher")
        chunk, channels = aud.shape if len(aud.shape) > 1 else (aud.shape[0], 1)
        self.chunk = chunk if self.chunk == -1 else self.chunk
        self.channels = channels if self.channels == -1 else self.channels
        if 0 < self.chunk != chunk or 0 < self.channels != channels:
            raise ValueError("Incorrect audio shape for publisher")
        aud = np.require(aud, dtype=np.float32, requirements="C")

        aud_str = json.dumps(
            (chunk, channels, rate, aud),
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        ).encode()
        aud_header = "{timestamp:" + str(time.time()) + "}"
        if hasattr(self, "_thread_local_storage"):
            socket = self._get_socket()
        else:
            socket = self._socket
        socket.send_multipart([self._topic, aud_header.encode(), aud_str])


@Publishers.register("Properties", "zeromq")
class ZeroMQPropertiesPublisher(ZeroMQPublisher):

    def __init__(self, name, out_topic, **kwargs):
        super().__init__(name, out_topic, **kwargs)
        raise NotImplementedError



"""
Encoder and Decoder for Pint Quantity Data via Wrapyfi.

This script provides mechanisms to encode and decode Pint Quantity Data using Wrapyfi.
It utilizes base64 encoding and JSON serialization.

The script contains a class, `PintData`, registered as a plugin to manage the
conversion of Pint Quantity data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - Pint: A package to define, operate and manipulate physical quantities (refer to https://pint.readthedocs.io/en/stable/ for installation instructions)
        Note: If Pint is not available, HAVE_PINT will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install pint``  # Basic installation of Pint
"""

import base64
import json

from wrapyfi.utils import *

try:
    from pint import Quantity

    HAVE_PINT = True
except ImportError:
    HAVE_PINT = False


@PluginRegistrar.register(types=None if not HAVE_PINT else Quantity.__mro__[:-1])
class PintData(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the PintData plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode Pint Quantity data into a base64 ASCII string.

        :param obj: pint.Quantity: The Pint Quantity data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, and object type
        """
        if isinstance(obj, Quantity):
            obj_type = "Quantity"
            obj_data = json.dumps(
                {"magnitude": obj.magnitude, "units": str(obj.units)}
            ).encode("ascii")
            obj_data = base64.b64encode(obj_data).decode("ascii")
            return True, dict(
                __wrapyfi__=(str(self.__class__.__name__), obj_data, obj_type)
            )
        else:
            # TypeError("Unknown object type: {}".format(obj_type))
            return False, {}

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into Pint Quantity data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string and object type
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, pint.Quantity]: A tuple containing:
            - bool: Indicating that the decoding was successful
            - pint.Quantity: The decoded Pint Quantity data
        """
        obj_data = base64.b64decode(obj_full[1].encode("ascii")).decode("ascii")
        obj_data = json.loads(obj_data)
        obj_type = obj_full[2]
        if obj_type == "Quantity":
            from pint import UnitRegistry

            ureg = UnitRegistry()
            obj = Quantity(
                obj_data["magnitude"], ureg.parse_expression(obj_data["units"])
            )
            return True, obj
        else:
            # TypeError("Unknown object type: {}".format(obj_type))
            return False, {}



import logging
import json
import queue
import os
from typing import Optional

import numpy as np
import cv2
import zmq

from wrapyfi.connect.clients import Client, Clients
from wrapyfi.middlewares.zeromq import ZeroMQMiddlewareReqRep
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook

SOCKET_IP = os.environ.get("WRAPYFI_ZEROMQ_SOCKET_IP", "127.0.0.1")
SOCKET_REP_PORT = int(os.environ.get("WRAPYFI_ZEROMQ_SOCKET_REP_PORT", 5558))
WATCHDOG_POLL_REPEAT = None


class ZeroMQClient(Client):
    def __init__(
        self,
        name,
        in_topic,
        carrier="tcp",
        socket_ip: str = SOCKET_IP,
        socket_rep_port: int = SOCKET_REP_PORT,
        zeromq_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the client.

        :param name: str: Name of the client
        :param out_topic: str: Topics are not supported for the REQ/REP pattern in ZeroMQ. Any given topic is ignored
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param zeromq_kwargs: dict: Additional kwargs for the ZeroMQ middleware
        :param kwargs: dict: Additional kwargs for the client
        """
        if in_topic != "":
            logging.warning(
                f"[ZeroMQ] ZeroMQ does not support topics for the REQ/REP pattern. Topic {in_topic} removed"
            )
            in_topic = ""
        if carrier or carrier != "tcp":
            logging.warning(
                "[ZeroMQ] ZeroMQ does not support other carriers than TCP for REQ/REP pattern. Using TCP."
            )
            carrier = "tcp"
        super().__init__(name, in_topic, carrier=carrier, **kwargs)

        self.socket_address = f"{carrier}://{socket_ip}:{socket_rep_port}"

        ZeroMQMiddlewareReqRep.activate(**zeromq_kwargs or {})

    def close(self):
        """
        Close the subscriber.
        """
        if hasattr(self, "_socket") and self._socket:
            if self._socket is not None:
                self._socket.close()

    def __del__(self):
        self.close()


@Clients.register("NativeObject", "zeromq")
class ZeroMQNativeObjectClient(ZeroMQClient):
    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "tcp",
        serializer_kwargs: Optional[dict] = None,
        deserializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Specific client for handling native Python objects, serializing them to JSON strings for transmission.

        :param name: str: Name of the client
        :param in_topic: str: Topics are not supported for the REQ/REP pattern in ZeroMQ. Any given topic is ignored
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for REQ/REP pattern. Default is 'tcp'
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, carrier=carrier, **kwargs)

        self._plugin_encoder = JsonEncoder
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self._queue = queue.Queue(maxsize=1)

        self.establish()

    def establish(self, **kwargs):
        """
        Establish the connection to the server.
        """
        self._socket = zmq.Context().instance().socket(zmq.REQ)
        for socket_property in ZeroMQMiddlewareReqRep().zeromq_kwargs.items():
            if isinstance(socket_property[1], str):
                self._socket.setsockopt_string(
                    getattr(zmq, socket_property[0]), socket_property[1]
                )
            else:
                self._socket.setsockopt(
                    getattr(zmq, socket_property[0]), socket_property[1]
                )
        self._socket.connect(self.socket_address)

        self.established = True

    def request(self, *args, **kwargs):
        """
        Serialize the provided Python objects to JSON strings, send a request to the server, and await a reply.
        The method uses the configured JSON encoder for serialization before sending the resultant string to the server.

        :param args: tuple: Arguments to be sent to the server
        :param kwargs: dict: Keyword arguments to be sent to the server
        :return: Any: The Python object received from the server, deserialized using the configured JSON decoder hook
        """
        try:
            self._request(*args, **kwargs)
        except zmq.ZMQError as e:
            logging.error("[ZeroMQ] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to serialize the request, send it to the server, and receive the reply.

        :param args: tuple: Arguments to be serialized and sent
        :param kwargs: dict: Keyword arguments to be serialized and sent
        """
        args_str = json.dumps(
            [args, kwargs], cls=self._plugin_encoder, **self._serializer_kwargs
        )
        self._socket.send_string(args_str)

        obj_str = self._socket.recv_string()
        obj = json.loads(
            obj_str, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs
        )
        self._queue.put(obj, block=False)

    def _await_reply(self):
        """
        Internal method to retrieve the reply from the server from the queue and return it.

        :return: Any: The Python object received from the server, deserialized using the configured JSON decoder hook
        """
        try:
            reply = self._queue.get(block=True)
            return reply
        except queue.Empty:
            logging.warning(
                f"[ZeroMQ] Discarding data because queue is empty. "
                f"This happened due to bad synchronization in {self.__class__.__name__}"
            )
            return None


@Clients.register("Image", "zeromq")
class ZeroMQImageClient(ZeroMQNativeObjectClient):
    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "tcp",
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        jpg: bool = False,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The Image client using the ZeroMQ message construct parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Topics are not supported for the REQ/REP pattern in ZeroMQ. Any given topic is ignored
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be compressed to JPG before sending. Default is False
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(
            name,
            in_topic,
            carrier=carrier,
            serializer_kwargs=serializer_kwargs,
            **kwargs,
        )
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._type = np.float32 if self.fp else np.uint8

    def _request(self, *args, **kwargs):
        """
        Internal method to serialize the request, send it to the server, and receive the reply.

        :param args: tuple: Arguments to be serialized and sent
        :param kwargs: dict: Keyword arguments to be serialized and sent
        """
        args_str = json.dumps(
            [args, kwargs], cls=self._plugin_encoder, **self._serializer_kwargs
        )
        self._socket.send_string(args_str)

        if self.jpg:
            reply_bytes = self._socket.recv()
            reply_img = cv2.imdecode(
                np.frombuffer(reply_bytes, np.uint8), cv2.IMREAD_ANYCOLOR
            )
        else:
            reply_str = self._socket.recv_string()
            reply_img_list = json.loads(reply_str)
            reply_img = np.array(reply_img_list["img"], dtype=self._type)
        self._queue.put(reply_img, block=False)

    def _await_reply(self):
        """
        Internal method to retrieve the reply from the server from the queue and return it.

        :return: np.ndarray: The image received from the server as a NumPy array
        """
        try:
            img = self._queue.get(block=True)
            height, width, channels = img.shape
            if (
                0 < self.width != width
                or 0 < self.height != height
                or img.size != height * width * (3 if self.rgb else 1)
            ):
                raise ValueError("Incorrect image shape for subscriber")
            return img
        except queue.Empty:
            logging.warning(
                f"[ZeroMQ] Discarding data because queue is empty. "
                f"This happened due to bad synchronization in {self.__class__.__name__}"
            )
            return None


@Clients.register("AudioChunk", "zeromq")
class ZeroMQAudioChunkClient(ZeroMQNativeObjectClient):
    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "tcp",
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The AudioChunk client using the ZeroMQ message construct parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Topics are not supported for the REQ/REP pattern in ZeroMQ. Any given topic is ignored
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be compressed to JPG before sending. Default is False
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(
            name,
            in_topic,
            carrier=carrier,
            serializer_kwargs=serializer_kwargs,
            **kwargs,
        )
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    def _request(self, *args, **kwargs):
        """
        Internal method to serialize the request, send it to the server, and receive the reply.

        :param args: tuple: Arguments to be serialized and sent
        :param kwargs: dict: Keyword arguments to be serialized and sent
        """
        args_str = json.dumps(
            [args, kwargs], cls=self._plugin_encoder, **self._serializer_kwargs
        )
        self._socket.send_string(args_str)

        reply_str = self._socket.recv_string()
        reply_aud_list = json.loads(reply_str)
        chunk, channels, rate, aud = reply_aud_list["aud"]
        reply_aud = np.array(aud, dtype=np.float32)
        self._queue.put((chunk, channels, rate, reply_aud), block=False)

    def _await_reply(self):
        """
        Internal method to retrieve the reply from the server from the queue and return it.

        :return: Tuple[np.ndarray, int]: Audio chunk received formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        try:
            chunk, channels, rate, aud = self._queue.get(block=True)
            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for listener")
            if (
                0 < self.chunk != chunk
                or self.channels != channels
                or aud.size != chunk * channels
            ):
                raise ValueError("Incorrect audio shape for listener")
            return aud, rate
        except queue.Empty:
            logging.warning(
                f"[ZeroMQ] Discarding data because queue is empty. "
                f"This happened due to bad synchronization in {self.__class__.__name__}"
            )
            return None



import logging
import sys
import json
import time
import os
import importlib
from typing import Optional, Tuple

import numpy as np
import cv2
import rclpy
from rclpy.node import Node
import std_msgs.msg
import sensor_msgs.msg

from wrapyfi.connect.publishers import Publisher, Publishers, PublisherWatchDog
from wrapyfi.middlewares.ros2 import ROS2Middleware
from wrapyfi.encoders import JsonEncoder


QUEUE_SIZE = int(os.environ.get("WRAPYFI_ROS2_QUEUE_SIZE", 5))
WATCHDOG_POLL_REPEAT = None


class ROS2Publisher(Publisher, Node):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        queue_size: int = QUEUE_SIZE,
        ros2_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the publisher.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        :param ros2_kwargs: dict: Additional kwargs for the ROS 2 middleware
        :param kwargs: dict: Additional kwargs for the publisher
        """
        carrier = "tcp"
        if "carrier" in kwargs and kwargs["carrier"] not in ["", None]:
            logging.warning(
                "[ROS 2] ROS 2 currently does not support explicit carrier setting for PUB/SUB pattern. Using TCP."
            )
        if "carrier" in kwargs:
            del kwargs["carrier"]
        ROS2Middleware.activate(**ros2_kwargs or {})
        Publisher.__init__(
            self, name, out_topic, carrier=carrier, should_wait=should_wait, **kwargs
        )
        Node.__init__(self, name + str(hex(id(self))))

        self.queue_size = queue_size

    def await_connection(
        self, publisher, out_topic: Optional[str] = None, repeats: Optional[int] = None
    ):
        """
        Wait for at least one subscriber to connect to the publisher.

        :param publisher: rclpy.publisher.Publisher: Publisher to await connection to
        :param out_topic: str: Name of the output topic
        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        connected = False
        if out_topic is None:
            out_topic = self.out_topic
        logging.info(f"[ROS 2] Waiting for topic subscriber: {out_topic}")
        if repeats is None:
            if self.should_wait:
                repeats = -1
            else:
                repeats = 1
            while repeats > 0 or repeats <= -1:
                repeats -= 1
                connected = publisher.get_subscription_count() > 0
                if connected:
                    break
                time.sleep(0.02)
        logging.info(f"[ROS 2] Topic subscriber connected: {out_topic}")
        return connected

    def close(self):
        """
        Close the publisher
        """
        if hasattr(self, "_publisher") and self._publisher:
            if self._publisher is not None:
                self.destroy_node()

    def __del__(self):
        self.close()


@Publishers.register("NativeObject", "ros2")
class ROS2NativeObjectPublisher(ROS2Publisher):

    def __init__(
        self,
        name,
        out_topic: str,
        should_wait: bool = True,
        queue_size: int = QUEUE_SIZE,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The NativeObject publisher using the ROS 2 String message assuming a combination of python native objects
        and numpy arrays as input. Serializes the data (including plugins) using the encoder and sends it as a string.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(
            name, out_topic, should_wait=should_wait, queue_size=queue_size, **kwargs
        )
        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        self._publisher = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        self._publisher = self.create_publisher(
            std_msgs.msg.String, self.out_topic, qos_profile=self.queue_size
        )
        established = self.await_connection(self._publisher, repeats=repeats)
        return self.check_establishment(established)

    def publish(self, obj):
        """
        Publish the object to the middleware

        :param obj: object: Object to publish
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)
        obj_str = json.dumps(
            obj,
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        )
        obj_str_msg = std_msgs.msg.String()
        obj_str_msg.data = obj_str
        self._publisher.publish(obj_str_msg)


@Publishers.register("Image", "ros2")
class ROS2ImagePublisher(ROS2Publisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        queue_size: int = QUEUE_SIZE,
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        jpg: bool = False,
        **kwargs,
    ):
        """
        The ImagePublisher using the ROS 2 Image message assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        :param width: int: Width of the image. Default is -1 meaning that the width is not fixed
        :param height: int: Height of the image. Default is -1 meaning that the height is not fixed
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be compressed as JPG. Default is False
        """
        super().__init__(
            name, out_topic, should_wait=should_wait, queue_size=queue_size, **kwargs
        )
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        if self.fp:
            self._encoding = "32FC3" if self.rgb else "32FC1"
            self._type = np.float32
        else:
            self._encoding = "bgr8" if self.rgb else "mono8"
            self._type = np.uint8
        if self.jpg:
            self._encoding = "jpeg"
            self._type = np.uint8

        self._publisher = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        if self.jpg:
            self._publisher = self.create_publisher(
                sensor_msgs.msg.CompressedImage,
                self.out_topic,
                qos_profile=self.queue_size,
            )
        else:
            self._publisher = self.create_publisher(
                sensor_msgs.msg.Image, self.out_topic, qos_profile=self.queue_size
            )
        established = self.await_connection(self._publisher)
        return self.check_establishment(established)

    def publish(self, img: np.ndarray):
        """
        Publish the image to the middleware.

        :param img: np.ndarray: Image to publish formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if img is None:
            return

        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        if (
            0 < self.width != img.shape[1]
            or 0 < self.height != img.shape[0]
            or not (
                (img.ndim == 2 and not self.rgb)
                or (img.ndim == 3 and self.rgb and img.shape[2] == 3)
            )
        ):
            raise ValueError("Incorrect image shape for publisher")
        img = np.require(img, dtype=self._type, requirements="C")

        if self.jpg:
            img_msg = sensor_msgs.msg.CompressedImage()
            img_msg.header.stamp = rclpy.clock.Clock().now().to_msg()
            img_msg.format = "jpeg"
            img_msg.data = np.array(cv2.imencode(".jpg", img)[1]).tobytes()
        else:
            img_msg = sensor_msgs.msg.Image()
            img_msg.header.stamp = self.get_clock().now().to_msg()
            img_msg.height = img.shape[0]
            img_msg.width = img.shape[1]
            img_msg.encoding = self._encoding
            img_msg.is_bigendian = img.dtype.byteorder == ">" or (
                img.dtype.byteorder == "=" and sys.byteorder == "big"
            )
            img_msg.step = img.strides[0]
            img_msg.data = img.tobytes()
        self._publisher.publish(img_msg)


@Publishers.register("AudioChunk", "ros2")
class ROS2AudioChunkPublisher(ROS2Publisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        queue_size: int = QUEUE_SIZE,
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        **kwargs,
    ):
        """
        The AudioChunkPublisher using the ROS 2 Audio message assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        :param channels: int: Number of channels. Default is 1
        :param rate: int: Sampling rate. Default is 44100
        :param chunk: int: Chunk size. Default is -1 meaning that the chunk size is not fixed
        """
        super().__init__(
            name, out_topic, should_wait=should_wait, queue_size=queue_size, **kwargs
        )
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._publisher = self._sound_msg = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        try:
            from wrapyfi_ros2_interfaces.msg import ROS2AudioMessage
        except ImportError:
            import wrapyfi

            logging.error(
                "[ROS 2] Could not import ROS2AudioMessage. "
                "Make sure the ROS 2 services in wrapyfi_extensions/wrapyfi_ros2_interfaces are compiled. "
                "Refer to the documentation for more information: \n"
                + wrapyfi.__doc__
                + "ros2_interfaces_lnk.html"
            )
            sys.exit(1)
        self._publisher = self.create_publisher(
            ROS2AudioMessage, self.out_topic, qos_profile=self.queue_size
        )
        self._sound_msg = ROS2AudioMessage()
        established = self.await_connection(self._publisher)
        return self.check_establishment(established)

    def publish(self, aud: Tuple[np.ndarray, int]):
        """
        Publish the audio chunk to the middleware.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        aud, rate = aud
        if aud is None:
            return
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for publisher")
        chunk, channels = aud.shape if len(aud.shape) > 1 else (aud.shape[0], 1)
        self.chunk = chunk if self.chunk == -1 else self.chunk
        self.channels = channels if self.channels == -1 else self.channels
        if 0 < self.chunk != chunk or 0 < self.channels != channels:
            raise ValueError("Incorrect audio shape for publisher")
        aud = np.require(aud, dtype=np.float32, requirements="C")

        aud_msg = self._sound_msg
        aud_msg.header.stamp = self.get_clock().now().to_msg()
        aud_msg.chunk_size = chunk
        aud_msg.channels = channels
        aud_msg.sample_rate = rate
        aud_msg.is_bigendian = aud.dtype.byteorder == ">" or (
            aud.dtype.byteorder == "=" and sys.byteorder == "big"
        )
        aud_msg.encoding = "S16BE" if aud_msg.is_bigendian else "S16LE"
        aud_msg.step = aud.strides[0]
        aud_msg.data = aud.tobytes()  # (aud * 32767.0).tobytes()
        self._publisher.publish(aud_msg)


@Publishers.register("Properties", "ros2")
class ROS2PropertiesPublisher(ROS2Publisher):
    def __init__(self, name, out_topic, **kwargs):
        super().__init__(name, out_topic, **kwargs)
        raise NotImplementedError


@Publishers.register("ROS2Message", "ros2")
class ROS2MessagePublisher(ROS2Publisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        queue_size: int = QUEUE_SIZE,
        **kwargs,
    ):
        """
        The ROS2MessagePublisher using the ROS 2 message type determined dynamically.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        """
        super().__init__(
            name, out_topic, should_wait=should_wait, queue_size=queue_size, **kwargs
        )
        self._publisher = None
        self._msg_type = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def get_message_type(self, msg):
        """
        Get the type of a specific message.

        :param msg: ROS 2 message object
        :return: type: The type of the provided message
        """
        return type(msg)

    def establish(self, msg, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection using the provided message to determine the type.

        :param msg: ROS2Message: Message to determine the type.
        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        self._msg_type = self.get_message_type(msg)

        self._publisher = self.create_publisher(
            self._msg_type, self.out_topic, qos_profile=self.queue_size
        )
        established = self.await_connection(self._publisher)
        return self.check_establishment(established)

    def publish(self, msg):
        """
        Publish the data to the middleware.

        :param msg: ROS2Message: Message to publish. This should be formatted according to the expected message type.
        """
        if not self._publisher:
            self.establish(msg)

        if not self.established:
            established = self.establish(msg, repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        self._publisher.publish(msg)
